#include <stdio.h>
#include <stdlib.h>
#include "driver/gpio.h"
#include "driver/adc.h"
#include "esp_adc_cal.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

#include "battery_adc.h"
#include "env_monitor.h"

#define DEFAULT_VREF    1100        // Use adc2_vref_to_gpio() to obtain a better estimate
#define ADC_EN_GPIO     25          // Enable measurement - GPIO25

static esp_adc_cal_characteristics_t *adc_chars;
static const adc_channel_t channel = ADC_CHANNEL_0;     // GPIO36
static const adc_atten_t atten = ADC_ATTEN_DB_0;
static const adc_unit_t unit = ADC_UNIT_1;

static RTC_DATA_ATTR bool is_batt_empty;

static float v_batt;
static bool is_adc_initialized;

void batt_adc_init() {
    // Configure ADC
    adc1_config_width(ADC_WIDTH_BIT_12);
    adc1_config_channel_atten(channel, atten);

    // Characterize ADC
    adc_chars = calloc(1, sizeof(esp_adc_cal_characteristics_t));
    esp_adc_cal_characterize(unit, atten, ADC_WIDTH_BIT_12, DEFAULT_VREF, adc_chars);

    // ADC_EN_GPIO init
    //gpio_reset_pin(ADC_EN_GPIO);
    gpio_set_direction(ADC_EN_GPIO, GPIO_MODE_OUTPUT);
    gpio_set_level(ADC_EN_GPIO, 0);

    is_adc_initialized = true;
}

bool batt_adc_is_initialized() {
    return is_adc_initialized;
}

static float batt_adc_process_measurement() {
    float ret = 0;

    uint32_t adc_reading = adc1_get_raw((adc1_channel_t)channel);
    uint32_t v_adc = esp_adc_cal_raw_to_voltage(adc_reading, adc_chars);
    ret = ((v_adc*(1500+470)) / (470.0*1000));

    printf("ADC raw: %d\n", adc_reading);
    printf("Vbatt: %f\n", ret);

    return ret;
}

esp_err_t batt_adc_measure() {
    // Enable measurement
    esp_err_t ret = gpio_set_level(ADC_EN_GPIO, 1);
    if (ret == ESP_OK) {
        ret = env_monitor_disable_tps_power_save_mode();

        if (ret == ESP_OK) {
            // Wait until C2 is charged
            vTaskDelay((1/portTICK_PERIOD_MS) + 1);

            v_batt = batt_adc_process_measurement();
        }
    }

    // Disable measurement
    gpio_set_level(ADC_EN_GPIO, 0);
    env_monitor_enable_tps_power_save_mode();

    return ret;
}

esp_err_t batt_adc_start_diagnostics() {
    // Enable measurement
    esp_err_t ret = gpio_set_level(ADC_EN_GPIO, 1);
    if (ret == ESP_OK) {
        ret = env_monitor_disable_tps_power_save_mode();

        if (ret == ESP_OK) {
            // Wait until C2 is charged
            vTaskDelay((1/portTICK_PERIOD_MS) + 1);
            
            ret = ESP_FAIL;
            float prev = 0;
            for (int i = 0; i <= 100; i++) {
                printf("ADC measure: %d\n", i);
                float voltage = batt_adc_process_measurement();

                if ((voltage < 3.0) || (voltage > 4.5)) {
                    ret = ESP_FAIL;
                    break;
                }
                else if ((i > 0) && (prev != voltage)) {
                    ret = ESP_OK;
                    break;
                }
                prev = voltage;
            }
        }
    }

    // Disable measurement
    gpio_set_level(ADC_EN_GPIO, 0);
    env_monitor_enable_tps_power_save_mode();

    return ret;
}

float batt_adc_get_vbatt() {
    return v_batt;
}

bool batt_adc_check_batt_state() {
    bool prev = is_batt_empty;

    if (v_batt < 3.3) {
        is_batt_empty = true;
    }
    else if (v_batt > 3.6) {
        is_batt_empty = false;
    }

    return (prev == is_batt_empty);
}

bool batt_adc_is_batt_empty() {
    return is_batt_empty;
}