#include "esp_wifi.h"
#include "esp_log.h"
#include "freertos/event_groups.h"
#include "firestore.h"
#include "string.h"
#include "sys/time.h"

#include "communication.h"
#include "rgb.h"

#define ESP_MAXIMUM_RETRY   5
#define ESP_WIFI_SSID       "put-your-wifi-ssid-here"
#define ESP_WIFI_PASS       "put-your-wifi-password-here"
#define TAG                 "COMMUNICATION"

/* The event group allows multiple bits for each event, but we only care about two events:
 * - we are connected to the AP with an IP
 * - we failed to connect after the maximum amount of retries */
#define WIFI_CONNECTED_BIT  BIT0
#define WIFI_FAIL_BIT       BIT1

/* FreeRTOS event group to signal when we are connected*/
static EventGroupHandle_t s_wifi_event_group;

static int s_retry_num = 0;

static bool is_wifi_initialized;

static void com_success() {
    rgb_light_ms(0, 5, 0, 500);
}

static void com_error() {
    rgb_light_ms(5, 0, 0, 500);
}

void com_event_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data) {
    if (event_base == WIFI_EVENT && event_id == WIFI_EVENT_STA_START) {
        esp_wifi_connect();
    } else if (event_base == WIFI_EVENT && event_id == WIFI_EVENT_STA_DISCONNECTED) {
        if (s_retry_num < ESP_MAXIMUM_RETRY) {
            esp_wifi_connect();
            s_retry_num++;
            ESP_LOGI(TAG, "retry to connect to the AP");
        } else {
            xEventGroupSetBits(s_wifi_event_group, WIFI_FAIL_BIT);
        }
        ESP_LOGI(TAG,"connect to the AP fail");
    } else if (event_base == IP_EVENT && event_id == IP_EVENT_STA_GOT_IP) {
        ip_event_got_ip_t* event = (ip_event_got_ip_t*) event_data;
        ESP_LOGI(TAG, "got ip:%s",
                 ip4addr_ntoa(&event->ip_info.ip));
        s_retry_num = 0;
        xEventGroupSetBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
    }
}

esp_err_t com_wifi_init_sta() {
    rgb_set_duty(5, 5, 0);

    s_wifi_event_group = xEventGroupCreate();

    // Initialize the underlying TCP/IP stack
    tcpip_adapter_init();

    ESP_ERROR_CHECK(esp_event_loop_create_default());

    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    ESP_ERROR_CHECK(esp_wifi_init(&cfg));

    ESP_ERROR_CHECK(esp_event_handler_register(WIFI_EVENT, ESP_EVENT_ANY_ID, &com_event_handler, NULL));
    ESP_ERROR_CHECK(esp_event_handler_register(IP_EVENT, IP_EVENT_STA_GOT_IP, &com_event_handler, NULL));

    wifi_config_t wifi_config = {
        .sta = {
            .ssid = ESP_WIFI_SSID,
            .password = ESP_WIFI_PASS,
	        .threshold.authmode = WIFI_AUTH_WPA2_PSK,
            .pmf_cfg = {
                .capable = true,
                .required = false
            },
        },
    };
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
    ESP_ERROR_CHECK(esp_wifi_start() );

    ESP_LOGI(TAG, "wifi_init_sta finished.");

    /* Waiting until either the connection is established (WIFI_CONNECTED_BIT) or connection failed for the maximum
     * number of re-tries (WIFI_FAIL_BIT). The bits are set by com_event_handler() (see above) */
    EventBits_t bits = xEventGroupWaitBits(s_wifi_event_group,
            WIFI_CONNECTED_BIT | WIFI_FAIL_BIT,
            pdFALSE,
            pdFALSE,
            portMAX_DELAY);

    esp_err_t ret = ESP_OK;

    /* xEventGroupWaitBits() returns the bits before the call returned, hence we can test which event actually
     * happened. */
    if (bits & WIFI_CONNECTED_BIT) {
        //ESP_LOGI(TAG, "connected to ap SSID:%s password:%s", ESP_WIFI_SSID, ESP_WIFI_PASS);
        ESP_LOGI(TAG, "connected to ap SSID:%s", ESP_WIFI_SSID);
    } else if (bits & WIFI_FAIL_BIT) {
        //ESP_LOGI(TAG, "Failed to connect to SSID:%s, password:%s", ESP_WIFI_SSID, ESP_WIFI_PASS);
        ESP_LOGI(TAG, "Failed to connect to SSID:%s", ESP_WIFI_SSID);
        ret = ESP_FAIL;
    } else {
        ESP_LOGE(TAG, "UNEXPECTED EVENT");
        ret = ESP_FAIL;
    }

    ESP_ERROR_CHECK(esp_event_handler_unregister(IP_EVENT, IP_EVENT_STA_GOT_IP, &com_event_handler));
    ESP_ERROR_CHECK(esp_event_handler_unregister(WIFI_EVENT, ESP_EVENT_ANY_ID, &com_event_handler));
    vEventGroupDelete(s_wifi_event_group);

    if (ret == ESP_OK) {
        rgb_stop();
        is_wifi_initialized = true;
    }
    else {
        com_error();
    }


    return ret;
}

bool com_is_wifi_initialized() {
    return is_wifi_initialized;
}

void com_get_firestore_timestamp_value(long timestamp_value, char *timestamp_str) {
    struct tm timeinfo;
    gmtime_r(&timestamp_value, &timeinfo);
    sprintf(timestamp_str, "%d-%02d-%02dT%02d:%02d:%02dZ",
                            (timeinfo.tm_year+1900),
                            (timeinfo.tm_mon+1),
                            timeinfo.tm_mday,
                            timeinfo.tm_hour,
                            timeinfo.tm_min,
                            (timeinfo.tm_sec%60));
}

/* Firestore get Document */
firestore_err_t com_get_document(char *collection_id, char *document_id, char **document) {
    uint32_t document_len;

    firestore_err_t ret = firestore_get_document(collection_id, document_id, document, &document_len);

    if (ret == FIRESTORE_OK) {
        com_success();
        ESP_LOGI(TAG, "Document length: %d", document_len);
        ESP_LOGI(TAG, "Document content:\r\n%.*s", document_len, *document);
    }
    else {
        com_error();
        ESP_LOGE(TAG, "Couldn't get document, error: %d", ret);
    }

    return ret;
}

/* Firestore add Document */
firestore_err_t com_add_document(char *collection_id, char *document_id, char *document) {
    uint32_t u32DocumentLen;

    firestore_err_t ret = firestore_add_document(collection_id, document_id, document, &u32DocumentLen);

    if (ret == FIRESTORE_OK) {
        com_success();
        ESP_LOGI(TAG, "Added document length: %d", u32DocumentLen);
        ESP_LOGI(TAG, "Added document content:\r\n%.*s", u32DocumentLen, document);
    }
    else {
        com_error();
        ESP_LOGE(TAG, "Couldn't add document, error: %d", ret);
    }

    return ret;
}

/* Firestore update Document */
firestore_err_t com_update_document(char *collection_id, char *document_id, char *document) {
    uint32_t u32DocumentLen;

    firestore_err_t ret = firestore_update_document(collection_id, document_id, document, &u32DocumentLen);

    if (ret == FIRESTORE_OK) {
        com_success();
        ESP_LOGI(TAG, "Updated document length: %d", u32DocumentLen);
        ESP_LOGI(TAG, "Updated document content:\r\n%.*s", u32DocumentLen, document);
    }
    else {
        com_error();
        ESP_LOGE(TAG, "Couldn't update document, error: %d", ret);
    }

    return ret;
}