#include "sensors.h"
#include "string.h"
#include "scd30.h"
#include "esp_log.h"

#define I2C_SDA_GPIO        23
#define I2C_SCL_GPIO        22
#define I2C_FREQ_HZ         100000              // I2C SCL frequency
#define WRITE_BIT           I2C_MASTER_WRITE    // I2C master write
#define ACK_CHECK_EN        0x1                 // I2C master will check ack from slave
#define READ_BIT            I2C_MASTER_READ     // I2C master read
#define ACK_VAL             0x0                 // I2C ack value
#define NACK_VAL            0x1                 // I2C nack value

#define TAG                 "SENSORS"

static bool i2c0_initialized;
static bool i2c1_initialized;

esp_err_t sens_i2c0_init() {
    esp_err_t ret;

    i2c_config_t conf = {
        .mode = I2C_MODE_MASTER,
        .sda_io_num = I2C_SDA_GPIO,
        .sda_pullup_en = GPIO_PULLUP_ENABLE,
        .scl_io_num = I2C_SCL_GPIO,
        .scl_pullup_en = GPIO_PULLUP_ENABLE,
        .master.clk_speed = I2C_FREQ_HZ,
    };

    i2c_param_config(I2C_NUM_0, &conf);

    ret = i2c_driver_install(I2C_NUM_0, conf.mode, 0, 0, 0);

    if (ret == ESP_OK) {
        i2c0_initialized = true;
    }
    else {
        ESP_LOGE(TAG, "I2C0 init error");
    }

    return ret;
}

esp_err_t sens_i2c1_init() {
    esp_err_t ret;

    i2c_config_t conf = {
        .mode = I2C_MODE_MASTER,
        .sda_io_num = I2C_SCD30_SDA_GPIO,
        .sda_pullup_en = GPIO_PULLUP_ENABLE,
        .scl_io_num = I2C_SCD30_SCL_GPIO,
        .scl_pullup_en = GPIO_PULLUP_ENABLE,
        .master.clk_speed = I2C_FREQ_HZ,
    };

    i2c_param_config(I2C_NUM_1, &conf);

    i2c_set_timeout(I2C_NUM_1, 0xFFFFF);

    ret = i2c_driver_install(I2C_NUM_1, conf.mode, 0, 0, 0);

    if (ret == ESP_OK) {
        i2c1_initialized = true;
    }
    else {
        ESP_LOGE(TAG, "I2C1 init error");
    }

    return ret;
}

bool sens_get_i2c0_initialized() {
    return i2c0_initialized;
}

bool sens_get_i2c1_initialized() {
    return i2c1_initialized;
}

esp_err_t sens_i2c_write(i2c_port_t i2c_num, uint8_t addr, uint8_t *data, uint16_t data_len) {
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, (addr << 1) | WRITE_BIT, ACK_CHECK_EN);

    for (int i = 0; i < data_len; i++) {
        i2c_master_write_byte(cmd, data[i], ACK_CHECK_EN);
    }

    i2c_master_stop(cmd);
    esp_err_t ret = i2c_master_cmd_begin(i2c_num, cmd, 1000 / portTICK_RATE_MS);
    i2c_cmd_link_delete(cmd);
    return ret;
}

esp_err_t sens_i2c_write_byte_to_reg(i2c_port_t i2c_num, uint8_t sens_addr, uint8_t sens_reg_addr, uint8_t data) {
    uint8_t buff[2];
    buff[0] = sens_reg_addr;
    buff[1] = data;
    esp_err_t ret = sens_i2c_write(i2c_num, sens_addr, buff, 2);
    return ret;
}

esp_err_t sens_i2c_read(i2c_port_t i2c_num, uint8_t addr, uint8_t *data, uint16_t data_len) {
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, (addr << 1) | READ_BIT, ACK_CHECK_EN);

    for (int i = 0; i < data_len; i++) {
        if (i == data_len-1) {
            i2c_master_read_byte(cmd, &data[i], NACK_VAL);    
        }
        else {
            i2c_master_read_byte(cmd, &data[i], ACK_VAL);
        }
    }

    i2c_master_stop(cmd);
    esp_err_t ret = i2c_master_cmd_begin(i2c_num, cmd, 1000 / portTICK_RATE_MS);
    i2c_cmd_link_delete(cmd);
    return ret;
}

esp_err_t sens_i2c_read_from_reg(i2c_port_t i2c_num, uint8_t sens_addr, uint8_t sens_reg_addr, uint8_t *data, uint16_t data_len) {
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, sens_addr << 1 | WRITE_BIT, ACK_CHECK_EN);
    i2c_master_write_byte(cmd, sens_reg_addr, ACK_CHECK_EN);
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, sens_addr << 1 | READ_BIT, ACK_CHECK_EN);

    for (int i = 0; i < data_len; i++) {
        if (i == data_len-1) {
            i2c_master_read_byte(cmd, &data[i], NACK_VAL);    
        }
        else {
            i2c_master_read_byte(cmd, &data[i], ACK_VAL);
        }
    }

    i2c_master_stop(cmd);
    esp_err_t ret = i2c_master_cmd_begin(i2c_num, cmd, 1000 / portTICK_RATE_MS);
    i2c_cmd_link_delete(cmd);
    return ret;
}

int16_t sens_scd30_init() {
    int16_t ret = NO_ERROR;

    gpio_set_direction(SCD30_PWREN_GPIO, GPIO_MODE_OUTPUT);
    gpio_set_level(SCD30_PWREN_GPIO, 0);
    vTaskDelay(100/portTICK_PERIOD_MS);

    uint16_t timeout = 0;
    while ((scd30_probe() != NO_ERROR) && (timeout < 100)) {
        timeout++;
        vTaskDelay((1/portTICK_PERIOD_MS) + 1);
    }

    if (timeout < 100) {
        printf("SCD30 sensor probing successful\n");
    }
    else {
        ret = ESP_FAIL;
    }

    if (ret == NO_ERROR) {
        ret = scd30_set_measurement_interval(2);
        if (ret == NO_ERROR) {
            sensirion_sleep_usec(20000u);
            ret = scd30_start_periodic_measurement(0);
        }
    }

    if (ret != NO_ERROR) {
        ESP_LOGE(TAG, "SCD30 init error");
    }

    return ret;
}

void sens_scd30_stop() {
    scd30_stop_periodic_measurement();
    gpio_set_level(SCD30_PWREN_GPIO, 1);

    gpio_config_t cfg;
    cfg.pin_bit_mask = BIT64(I2C_SCD30_SCL_GPIO),
    cfg.mode = GPIO_MODE_DISABLE,
    cfg.pull_up_en = false,
    cfg.pull_down_en = true,
    cfg.intr_type = GPIO_INTR_DISABLE,
    gpio_config(&cfg);

    cfg.pin_bit_mask = BIT64(I2C_SCD30_SDA_GPIO),
    cfg.mode = GPIO_MODE_DISABLE,
    cfg.pull_up_en = false,
    cfg.pull_down_en = true,
    cfg.intr_type = GPIO_INTR_DISABLE,
    gpio_config(&cfg);
}
